@@ -46,22 +46,25 @@
 #include "src/dbus.h"
 
 struct ap_state {
-	struct device *device;
+	struct netdev *netdev;
 	char *ssid;
 	int channel;
 	unsigned int ciphers;
+	enum ie_rsn_cipher_suite group_cipher;
 	uint32_t beacon_interval;
 	struct l_uintset *rates;
 	uint8_t pmk[32];
 	struct l_queue *frame_watch_ids;
 	uint32_t start_stop_cmd_id;
-	uint32_t netdev_watch_id;
+	uint8_t gtk[CRYPTO_MAX_GTK_LEN];
+	uint8_t gtk_index;
 
 	uint16_t last_aid;
 	struct l_queue *sta_states;
 
 	struct l_dbus_message *pending;
 	bool started : 1;
+	bool gtk_set : 1;
 };
 
 struct sta_state {
@@ -75,13 +78,13 @@ struct sta_state {
 	uint32_t assoc_resp_cmd_id;
 	struct ap_state *ap;
 	uint8_t *assoc_rsne;
-	size_t assoc_rsne_len;
 	struct eapol_sm *sm;
 	struct handshake_state *hs;
+	uint32_t gtk_query_cmd_id;
 };
 
 static struct l_genl_family *nl80211 = NULL;
-static uint32_t device_watch;
+static uint32_t netdev_watch;
 
 static void ap_sta_free(void *data)
 {
@@ -93,8 +96,14 @@ static void ap_sta_free(void *data)
 	if (sta->assoc_resp_cmd_id)
 		l_genl_family_cancel(nl80211, sta->assoc_resp_cmd_id);
 
-	eapol_sm_free(sta->sm);
-	handshake_state_free(sta->hs);
+	if (sta->gtk_query_cmd_id)
+		l_genl_family_cancel(nl80211, sta->gtk_query_cmd_id);
+
+	if (sta->sm)
+		eapol_sm_free(sta->sm);
+
+	if (sta->hs)
+		handshake_state_free(sta->hs);
 
 	l_free(sta);
 }
@@ -109,7 +118,7 @@ static void ap_frame_watch_remove(void *data, void *user_data)
 
 static void ap_reset(struct ap_state *ap)
 {
-	struct netdev *netdev = device_get_netdev(ap->device);
+	struct netdev *netdev = ap->netdev;
 
 	if (ap->pending)
 		dbus_pending_reply(&ap->pending,
@@ -125,14 +134,15 @@ static void ap_reset(struct ap_state *ap)
 	if (ap->start_stop_cmd_id)
 		l_genl_family_cancel(nl80211, ap->start_stop_cmd_id);
 
-	netdev_watch_remove(netdev, ap->netdev_watch_id);
-
 	l_queue_destroy(ap->sta_states, ap_sta_free);
 
 	if (ap->rates)
 		l_uintset_free(ap->rates);
 
 	ap->started = false;
+
+	l_dbus_property_changed(dbus_get_bus(), netdev_get_path(ap->netdev),
+						IWD_AP_INTERFACE, "Started");
 }
 
 static void ap_free(void *data)
@@ -147,10 +157,23 @@ static void ap_free(void *data)
 static void ap_del_station(struct sta_state *sta, uint16_t reason,
 				bool disassociate)
 {
-	netdev_del_station(device_get_netdev(sta->ap->device), sta->addr,
-				reason, disassociate);
+	netdev_del_station(sta->ap->netdev, sta->addr, reason, disassociate);
 	sta->associated = false;
 	sta->rsna = false;
+
+	if (sta->gtk_query_cmd_id) {
+		l_genl_family_cancel(nl80211, sta->gtk_query_cmd_id);
+		sta->gtk_query_cmd_id = 0;
+	}
+
+	if (sta->sm)
+		eapol_sm_free(sta->sm);
+
+	if (sta->hs)
+		handshake_state_free(sta->hs);
+
+	sta->hs = NULL;
+	sta->sm = NULL;
 }
 
 static bool ap_sta_match_addr(const void *a, const void *b)
@@ -163,7 +186,7 @@ static bool ap_sta_match_addr(const void *a, const void *b)
 static void ap_remove_sta(struct sta_state *sta)
 {
 	if (!l_queue_remove(sta->ap->sta_states, sta)) {
-		l_error("tried to remove station that doesnt exist");
+		l_error("tried to remove station that doesn't exist");
 		return;
 	}
 
@@ -196,7 +219,7 @@ static void ap_new_rsna(struct sta_state *sta)
 static void ap_drop_rsna(struct sta_state *sta)
 {
 	struct l_genl_msg *msg;
-	uint32_t ifindex = device_get_ifindex(sta->ap->device);
+	uint32_t ifindex = netdev_get_ifindex(sta->ap->netdev);
 	struct nl80211_sta_flag_update flags = {
 		.mask = (1 << NL80211_STA_FLAG_AUTHORIZED) |
 			(1 << NL80211_STA_FLAG_MFP),
@@ -227,7 +250,12 @@ static void ap_drop_rsna(struct sta_state *sta)
 		l_error("Issuing DEL_KEY failed");
 	}
 
-	handshake_state_free(sta->hs);
+	if (sta->sm)
+		eapol_sm_free(sta->sm);
+
+	if (sta->hs)
+		handshake_state_free(sta->hs);
+
 	sta->hs = NULL;
 	sta->sm = NULL;
 }
@@ -237,7 +265,7 @@ static void ap_set_rsn_info(struct ap_state *ap, struct ie_rsn_info *rsn)
 	memset(rsn, 0, sizeof(*rsn));
 	rsn->akm_suites = IE_RSN_AKM_SUITE_PSK;
 	rsn->pairwise_ciphers = ap->ciphers;
-	rsn->group_cipher = IE_RSN_CIPHER_SUITE_NO_GROUP_TRAFFIC;
+	rsn->group_cipher = ap->group_cipher;
 }
 
 /*
@@ -253,7 +281,7 @@ static size_t ap_build_beacon_pr_head(struct ap_state *ap,
 	struct mmpdu_header *mpdu = (void *) out_buf;
 	unsigned int len;
 	uint16_t capability = IE_BSS_CAP_ESS | IE_BSS_CAP_PRIVACY;
-	const uint8_t *bssid = device_get_address(ap->device);
+	const uint8_t *bssid = netdev_get_address(ap->netdev);
 	uint32_t minr, maxr, count, r;
 	uint8_t *rates;
 	struct ie_tlv_builder builder;
@@ -334,7 +362,7 @@ static uint32_t ap_send_mgmt_frame(struct ap_state *ap,
 					void *user_data)
 {
 	struct l_genl_msg *msg;
-	uint32_t ifindex = device_get_ifindex(ap->device);
+	uint32_t ifindex = netdev_get_ifindex(ap->netdev);
 	uint32_t id;
 	uint32_t ch_freq = scan_channel_to_freq(ap->channel, SCAN_BAND_2_4_GHZ);
 
@@ -368,60 +396,51 @@ static void ap_handshake_event(struct handshake_state *hs,
 		netdev_handshake_failed(hs, l_get_u16(event_data));
 		/* fall through */
 	case HANDSHAKE_EVENT_SETTING_KEYS_FAILED:
+		sta->sm = NULL;
 		ap_remove_sta(sta);
 	default:
 		break;
 	}
 }
 
-static void ap_associate_sta_cb(struct l_genl_msg *msg, void *user_data)
+static void ap_start_rsna(struct sta_state *sta, const uint8_t *gtk_rsc)
 {
-	struct sta_state *sta = user_data;
 	struct ap_state *ap = sta->ap;
-	struct netdev *netdev = device_get_netdev(sta->ap->device);
+	struct netdev *netdev = sta->ap->netdev;
 	const uint8_t *own_addr = netdev_get_address(netdev);
 	struct scan_bss bss;
 	struct ie_rsn_info rsn;
 	uint8_t bss_rsne[24];
 
-	if (l_genl_msg_get_error(msg) < 0) {
-		l_error("NEW_STATION/SET_STATION failed: %i",
-			l_genl_msg_get_error(msg));
-		return;
-	}
-
 	memset(&bss, 0, sizeof(bss));
 
 	ap_set_rsn_info(ap, &rsn);
 	/*
 	 * TODO: This assumes the length that ap_set_rsn_info() requires. If
 	 * ap_set_rsn_info() changes then this will need to be updated.
-	 * Alternatively, sta->assoc_rsne could be used instead, but not
-	 * how it sits currently. sta->assoc_rsne only includes the actual RSN
-	 * data, not the IE type or length in the header.
 	 */
 	ie_build_rsne(&rsn, bss_rsne);
 
-	if (memcmp(bss_rsne + 2, sta->assoc_rsne, sta->assoc_rsne_len)) {
-		l_error("RSNE from association does not match");
-		goto error;
-	}
-
 	/* this handshake setup assumes PSK network */
 	sta->hs = netdev_handshake_state_new(netdev);
 
 	handshake_state_set_event_func(sta->hs, ap_handshake_event, sta);
 	handshake_state_set_ssid(sta->hs, (void *)ap->ssid, strlen(ap->ssid));
-	/* ap_rsn/own_rsn can be set equal since the check above matched */
-	handshake_state_set_ap_rsn(sta->hs, bss_rsne);
-	handshake_state_set_own_rsn(sta->hs, bss_rsne);
+	handshake_state_set_authenticator(sta->hs, true);
+	handshake_state_set_authenticator_rsn(sta->hs, bss_rsne);
+	handshake_state_set_supplicant_rsn(sta->hs, sta->assoc_rsne);
 	handshake_state_set_pmk(sta->hs, ap->pmk, 32);
 	handshake_state_set_authenticator_address(sta->hs, own_addr);
 	handshake_state_set_supplicant_address(sta->hs, sta->addr);
 
+	if (gtk_rsc)
+		handshake_state_set_gtk(sta->hs, ap->gtk, ap->gtk_index,
+					gtk_rsc);
+
 	sta->sm = eapol_sm_new(sta->hs);
 	if (!sta->sm) {
 		handshake_state_free(sta->hs);
+		sta->hs = NULL;
 		l_error("could not create sm object");
 		goto error;
 	}
@@ -429,7 +448,222 @@ static void ap_associate_sta_cb(struct l_genl_msg *msg, void *user_data)
 	eapol_sm_set_listen_interval(sta->sm, sta->listen_interval);
 	eapol_sm_set_protocol_version(sta->sm, EAPOL_PROTOCOL_VERSION_2004);
 
-	eapol_register_authenticator(sta->sm);
+	eapol_register(sta->sm);
+
+	return;
+
+error:
+	ap_del_station(sta, MMPDU_REASON_CODE_UNSPECIFIED, true);
+}
+
+static void ap_gtk_query_cb(struct l_genl_msg *msg, void *user_data)
+{
+	struct sta_state *sta = user_data;
+	struct l_genl_attr attr, nested;
+	uint16_t type, len;
+	const void *data;
+
+	sta->gtk_query_cmd_id = 0;
+
+	if (l_genl_msg_get_error(msg) < 0 || !l_genl_attr_init(&attr, msg)) {
+		l_error("GET_KEY failed for the GTK: %i",
+			l_genl_msg_get_error(msg));
+		goto error;
+	}
+
+	while (l_genl_attr_next(&attr, &type, &len, &data)) {
+		if (type != NL80211_ATTR_KEY)
+			continue;
+
+		break;
+	}
+
+	if (type != NL80211_ATTR_KEY || !l_genl_attr_recurse(&attr, &nested)) {
+		l_error("Can't recurse into ATTR_KEY in GET_KEY reply");
+		goto error;
+	}
+
+	while (l_genl_attr_next(&nested, &type, &len, &data)) {
+		if (type != NL80211_KEY_SEQ)
+			continue;
+
+		break;
+	}
+
+	if (type != NL80211_KEY_SEQ) {
+		l_error("KEY_SEQ not returned in GET_KEY reply");
+		goto error;
+	}
+
+	if (len != 6) {
+		l_error("KEY_SEQ length != 6 in GET_KEY reply");
+		goto error;
+	}
+
+	ap_start_rsna(sta, data);
+	return;
+
+error:
+	ap_del_station(sta, MMPDU_REASON_CODE_UNSPECIFIED, true);
+}
+
+static struct l_genl_msg *ap_build_cmd_new_key(struct ap_state *ap,
+						uint32_t cipher, size_t key_len)
+{
+	uint32_t ifindex = netdev_get_ifindex(ap->netdev);
+	struct l_genl_msg *msg;
+
+	msg = l_genl_msg_new_sized(NL80211_CMD_NEW_KEY, 128);
+
+	l_genl_msg_append_attr(msg, NL80211_ATTR_IFINDEX, 4, &ifindex);
+	l_genl_msg_enter_nested(msg, NL80211_ATTR_KEY);
+	l_genl_msg_append_attr(msg, NL80211_KEY_IDX, 1, &ap->gtk_index);
+	l_genl_msg_append_attr(msg, NL80211_KEY_DATA, key_len, ap->gtk);
+	l_genl_msg_append_attr(msg, NL80211_KEY_CIPHER, 4, &cipher);
+	l_genl_msg_leave_nested(msg);
+
+	return msg;
+}
+
+static struct l_genl_msg *ap_build_cmd_set_key(struct ap_state *ap)
+{
+	uint32_t ifindex = netdev_get_ifindex(ap->netdev);
+	struct l_genl_msg *msg;
+
+	msg = l_genl_msg_new_sized(NL80211_CMD_SET_KEY, 128);
+
+	l_genl_msg_append_attr(msg, NL80211_ATTR_IFINDEX, 4, &ifindex);
+	l_genl_msg_enter_nested(msg, NL80211_ATTR_KEY);
+	l_genl_msg_append_attr(msg, NL80211_KEY_IDX, 1, &ap->gtk_index);
+	l_genl_msg_append_attr(msg, NL80211_KEY_DEFAULT, 0, NULL);
+	l_genl_msg_enter_nested(msg, NL80211_KEY_DEFAULT_TYPES);
+	l_genl_msg_append_attr(msg, NL80211_KEY_DEFAULT_TYPE_MULTICAST,
+				0, NULL);
+	l_genl_msg_leave_nested(msg);
+	l_genl_msg_leave_nested(msg);
+
+	return msg;
+}
+
+static struct l_genl_msg *ap_build_cmd_del_key(struct ap_state *ap)
+{
+	uint32_t ifindex = netdev_get_ifindex(ap->netdev);
+	struct l_genl_msg *msg;
+
+	msg = l_genl_msg_new_sized(NL80211_CMD_DEL_KEY, 128);
+
+	l_genl_msg_append_attr(msg, NL80211_ATTR_IFINDEX, 4, &ifindex);
+	l_genl_msg_enter_nested(msg, NL80211_ATTR_KEY);
+	l_genl_msg_append_attr(msg, NL80211_KEY_IDX, 1, &ap->gtk_index);
+	l_genl_msg_leave_nested(msg);
+
+	return msg;
+}
+
+static struct l_genl_msg *ap_build_cmd_get_key(struct ap_state *ap)
+{
+	uint32_t ifindex = netdev_get_ifindex(ap->netdev);
+	struct l_genl_msg *msg;
+
+	msg = l_genl_msg_new_sized(NL80211_CMD_GET_KEY, 128);
+
+	l_genl_msg_append_attr(msg, NL80211_ATTR_IFINDEX, 4, &ifindex);
+	l_genl_msg_append_attr(msg, NL80211_ATTR_KEY_IDX, 1, &ap->gtk_index);
+
+	return msg;
+}
+
+static void ap_gtk_op_cb(struct l_genl_msg *msg, void *user_data)
+{
+	if (l_genl_msg_get_error(msg) < 0) {
+		uint8_t cmd = l_genl_msg_get_command(msg);
+		const char *cmd_name =
+			cmd == NL80211_CMD_NEW_KEY ? "NEW_KEY" :
+			cmd == NL80211_CMD_SET_KEY ? "SET_KEY" :
+			"DEL_KEY";
+
+		l_error("%s failed for the GTK: %i",
+			cmd_name, l_genl_msg_get_error(msg));
+	}
+}
+
+static void ap_associate_sta_cb(struct l_genl_msg *msg, void *user_data)
+{
+	struct sta_state *sta = user_data;
+	struct ap_state *ap = sta->ap;
+
+	if (l_genl_msg_get_error(msg) < 0) {
+		l_error("NEW_STATION/SET_STATION failed: %i",
+			l_genl_msg_get_error(msg));
+		return;
+	}
+
+	/*
+	 * Set up the group key.  If this is our first STA then we have
+	 * to add the new GTK to the kernel.  In theory we should be
+	 * able to supply our own RSC (e.g. generated randomly) and use it
+	 * immediately for our 4-Way Handshake without querying the kernel.
+	 * However NL80211_CMD_NEW_KEY only lets us set the receive RSC --
+	 * the Rx PN for CCMP and the Rx IV for TKIP -- and the
+	 * transmit RSC always starts as all zeros.  There's effectively
+	 * no way to set the Tx RSC or query the Rx RSC through nl80211.
+	 * So we query the Tx RSC in both scenarios just in case some
+	 * driver/hardware uses a different initial Tx RSC.
+	 *
+	 * Optimally we would get called back by the EAPoL state machine
+	 * only when building the step 3 of 4 message to query the RSC as
+	 * late as possible but that would complicate EAPoL.
+	 */
+	if (ap->group_cipher != IE_RSN_CIPHER_SUITE_NO_GROUP_TRAFFIC &&
+			!ap->gtk_set) {
+		enum crypto_cipher group_cipher =
+			ie_rsn_cipher_suite_to_cipher(ap->group_cipher);
+		int gtk_len = crypto_cipher_key_len(group_cipher);
+
+		/*
+		 * Generate our GTK.  Not following the example derivation
+		 * method in 802.11-2016 section 12.7.1.4 because a simple
+		 * l_getrandom is just as good.
+		 */
+		l_getrandom(ap->gtk, gtk_len);
+		ap->gtk_index = 1;
+
+		msg = ap_build_cmd_new_key(ap, group_cipher, gtk_len);
+		if (!l_genl_family_send(nl80211, msg, ap_gtk_op_cb, NULL,
+					NULL)) {
+			l_genl_msg_unref(msg);
+			l_error("Issuing NEW_KEY failed");
+			goto error;
+		}
+
+		msg = ap_build_cmd_set_key(ap);
+		if (!l_genl_family_send(nl80211, msg, ap_gtk_op_cb, NULL,
+					NULL)) {
+			l_genl_msg_unref(msg);
+			l_error("Issuing SET_KEY failed");
+			goto error;
+		}
+
+		/*
+		 * Set the flag now because any new associating STA will
+		 * just use NL80211_CMD_GET_KEY from now.
+		 */
+		ap->gtk_set = true;
+	}
+
+	if (ap->group_cipher == IE_RSN_CIPHER_SUITE_NO_GROUP_TRAFFIC)
+		ap_start_rsna(sta, NULL);
+	else {
+		msg = ap_build_cmd_get_key(ap);
+		sta->gtk_query_cmd_id = l_genl_family_send(nl80211, msg,
+								ap_gtk_query_cb,
+								sta, NULL);
+		if (!sta->gtk_query_cmd_id) {
+			l_genl_msg_unref(msg);
+			l_error("Issuing GET_KEY failed");
+			goto error;
+		}
+	}
 
 	return;
 
@@ -440,7 +674,7 @@ error:
 static void ap_associate_sta(struct ap_state *ap, struct sta_state *sta)
 {
 	struct l_genl_msg *msg;
-	uint32_t ifindex = device_get_ifindex(ap->device);
+	uint32_t ifindex = netdev_get_ifindex(ap->netdev);
 	/*
 	 * This should hopefully work both with and without
 	 * NL80211_FEATURE_FULL_AP_CLIENT_STATE.
@@ -535,7 +769,7 @@ static void ap_fail_assoc_resp_cb(struct l_genl_msg *msg, void *user_data)
 		l_error("AP (Re)Association Response with an error status not "
 			"sent or not ACKed: %i", l_genl_msg_get_error(msg));
 	else
-		l_info("AP (Re)Association Response with an errror status "
+		l_info("AP (Re)Association Response with an error status "
 			"delivered OK");
 }
 
@@ -544,7 +778,7 @@ static uint32_t ap_assoc_resp(struct ap_state *ap, struct sta_state *sta,
 				enum mmpdu_reason_code status_code,
 				bool reassoc, l_genl_msg_func_t callback)
 {
-	const uint8_t *addr = device_get_address(ap->device);
+	const uint8_t *addr = netdev_get_address(ap->netdev);
 	uint8_t mpdu_buf[128];
 	struct mmpdu_header *mpdu = (void *) mpdu_buf;
 	struct mmpdu_association_response *resp;
@@ -643,7 +877,7 @@ static void ap_assoc_reassoc(struct sta_state *sta, bool reassoc,
 	struct ap_state *ap = sta->ap;
 	const char *ssid = NULL;
 	const uint8_t *rsn = NULL;
-	size_t ssid_len = 0, rsn_len = 0;
+	size_t ssid_len = 0;
 	struct l_uintset *rates = NULL;
 	struct ie_rsn_info rsn_info;
 	int err;
@@ -678,8 +912,7 @@ static void ap_assoc_reassoc(struct sta_state *sta, bool reassoc,
 				goto bad_frame;
 			}
 
-			rsn = (const uint8_t *) ie_tlv_iter_get_data(ies);
-			rsn_len = ie_tlv_iter_get_length(ies);
+			rsn = (const uint8_t *) ie_tlv_iter_get_data(ies) - 2;
 			break;
 		}
 
@@ -730,8 +963,7 @@ static void ap_assoc_reassoc(struct sta_state *sta, bool reassoc,
 	if (sta->assoc_rsne)
 		l_free(sta->assoc_rsne);
 
-	sta->assoc_rsne = l_memdup(rsn, rsn_len);
-	sta->assoc_rsne_len = rsn_len;
+	sta->assoc_rsne = l_memdup(rsn, rsn[1] + 2);
 
 	/* 802.11-2016 11.3.5.3 j) */
 	if (sta->rsna)
@@ -780,7 +1012,7 @@ static void ap_assoc_req_cb(struct netdev *netdev,
 	struct sta_state *sta;
 	const uint8_t *from = hdr->address_2;
 	const struct mmpdu_association_request *req = body;
-	const uint8_t *bssid = device_get_address(ap->device);
+	const uint8_t *bssid = netdev_get_address(ap->netdev);
 	struct ie_tlv_iter iter;
 
 	l_info("AP Association Request from %s", util_address_to_string(from));
@@ -814,7 +1046,7 @@ static void ap_reassoc_req_cb(struct netdev *netdev,
 	struct sta_state *sta;
 	const uint8_t *from = hdr->address_2;
 	const struct mmpdu_reassociation_request *req = body;
-	const uint8_t *bssid = device_get_address(ap->device);
+	const uint8_t *bssid = netdev_get_address(ap->netdev);
 	struct ie_tlv_iter iter;
 	int err;
 
@@ -871,7 +1103,7 @@ static void ap_probe_req_cb(struct netdev *netdev,
 	size_t ssid_len = 0, ssid_list_len = 0, len;
 	int dsss_channel = -1;
 	struct ie_tlv_iter iter;
-	const uint8_t *bssid = device_get_address(ap->device);
+	const uint8_t *bssid = netdev_get_address(ap->netdev);
 	bool match = false;
 	uint8_t resp[512];
 
@@ -960,7 +1192,7 @@ static void ap_disassoc_cb(struct netdev *netdev,
 	struct ap_state *ap = user_data;
 	struct sta_state *sta;
 	const struct mmpdu_disassociation *disassoc = body;
-	const uint8_t *bssid = device_get_address(ap->device);
+	const uint8_t *bssid = netdev_get_address(ap->netdev);
 
 	l_info("AP Disassociation from %s, reason %i",
 		util_address_to_string(hdr->address_2),
@@ -995,7 +1227,7 @@ static void ap_auth_reply_cb(struct l_genl_msg *msg, void *user_data)
 static void ap_auth_reply(struct ap_state *ap, const uint8_t *dest,
 				enum mmpdu_reason_code status_code)
 {
-	const uint8_t *addr = device_get_address(ap->device);
+	const uint8_t *addr = netdev_get_address(ap->netdev);
 	uint8_t mpdu_buf[64];
 	struct mmpdu_header *mpdu = (struct mmpdu_header *) mpdu_buf;
 	struct mmpdu_authentication *auth;
@@ -1030,7 +1262,7 @@ static void ap_auth_cb(struct netdev *netdev, const struct mmpdu_header *hdr,
 	struct ap_state *ap = user_data;
 	const struct mmpdu_authentication *auth = body;
 	const uint8_t *from = hdr->address_2;
-	const uint8_t *bssid = device_get_address(ap->device);
+	const uint8_t *bssid = netdev_get_address(ap->netdev);
 	struct sta_state *sta;
 
 	l_info("AP Authentication from %s", util_address_to_string(from));
@@ -1099,7 +1331,7 @@ static void ap_deauth_cb(struct netdev *netdev, const struct mmpdu_header *hdr,
 	struct ap_state *ap = user_data;
 	struct sta_state *sta;
 	const struct mmpdu_deauthentication *deauth = body;
-	const uint8_t *bssid = device_get_address(ap->device);
+	const uint8_t *bssid = netdev_get_address(ap->netdev);
 
 	l_info("AP Deauthentication from %s, reason %i",
 		util_address_to_string(hdr->address_2),
@@ -1119,20 +1351,6 @@ static void ap_deauth_cb(struct netdev *netdev, const struct mmpdu_header *hdr,
 	ap_sta_free(sta);
 }
 
-static void ap_netdev_notify(struct netdev *netdev,
-				enum netdev_watch_event event, void *user_data)
-{
-	struct ap_state *ap = user_data;
-
-	switch (event) {
-	case NETDEV_WATCH_EVENT_DOWN:
-		ap_reset(ap);
-		break;
-	default:
-		break;
-	}
-}
-
 static void ap_start_cb(struct l_genl_msg *msg, void *user_data)
 {
 	struct ap_state *ap = user_data;
@@ -1156,6 +1374,9 @@ static void ap_start_cb(struct l_genl_msg *msg, void *user_data)
 			l_dbus_message_new_method_return(ap->pending));
 
 	ap->started = true;
+
+	l_dbus_property_changed(dbus_get_bus(), netdev_get_path(ap->netdev),
+						IWD_AP_INTERFACE, "Started");
 }
 
 static struct l_genl_msg *ap_build_cmd_start_ap(struct ap_state *ap)
@@ -1166,8 +1387,8 @@ static struct l_genl_msg *ap_build_cmd_start_ap(struct ap_state *ap)
 	size_t head_len, tail_len;
 
 	uint32_t dtim_period = 3;
-	uint32_t ifindex = device_get_ifindex(ap->device);
-	struct wiphy *wiphy = device_get_wiphy(ap->device);
+	uint32_t ifindex = netdev_get_ifindex(ap->netdev);
+	struct wiphy *wiphy = netdev_get_wiphy(ap->netdev);
 	uint32_t hidden_ssid = NL80211_HIDDEN_SSID_NOT_IN_USE;
 	uint32_t nl_ciphers = ie_rsn_cipher_suite_to_cipher(ap->ciphers);
 	uint32_t nl_akm = CRYPTO_AKM_PSK;
@@ -1228,8 +1449,8 @@ static struct l_genl_msg *ap_build_cmd_start_ap(struct ap_state *ap)
 static int ap_start(struct ap_state *ap, const char *ssid, const char *psk,
 		struct l_dbus_message *message)
 {
-	struct netdev *netdev = device_get_netdev(ap->device);
-	struct wiphy *wiphy = device_get_wiphy(ap->device);
+	struct netdev *netdev = ap->netdev;
+	struct wiphy *wiphy = netdev_get_wiphy(netdev);
 	struct l_genl_msg *cmd;
 	const struct l_queue_entry *entry;
 	uint32_t id;
@@ -1239,6 +1460,7 @@ static int ap_start(struct ap_state *ap, const char *ssid, const char *psk,
 	ap->channel = 6;
 	/* TODO: Add all ciphers supported by wiphy */
 	ap->ciphers = wiphy_select_cipher(wiphy, 0xffff);
+	ap->group_cipher = wiphy_select_cipher(wiphy, 0xffff);
 	ap->beacon_interval = 100;
 	/* TODO: Use actual supported rates */
 	ap->rates = l_uintset_new(200);
@@ -1298,8 +1520,6 @@ static int ap_start(struct ap_state *ap, const char *ssid, const char *psk,
 		goto error;
 	}
 
-	ap->netdev_watch_id = netdev_watch_add(netdev, ap_netdev_notify, ap);
-
 	ap->pending = l_dbus_message_ref(message);
 
 	return 0;
@@ -1336,7 +1556,7 @@ end:
 static struct l_genl_msg *ap_build_cmd_stop_ap(struct ap_state *ap)
 {
 	struct l_genl_msg *cmd;
-	uint32_t ifindex = device_get_ifindex(ap->device);
+	uint32_t ifindex = netdev_get_ifindex(ap->netdev);
 
 	cmd = l_genl_msg_new_sized(NL80211_CMD_STOP_AP, 16);
 	l_genl_msg_append_attr(cmd, NL80211_ATTR_IFINDEX, 4, &ifindex);
@@ -1362,6 +1582,19 @@ static int ap_stop(struct ap_state *ap, struct l_dbus_message *message)
 		return -EIO;
 	}
 
+	if (ap->gtk_set) {
+		struct l_genl_msg *msg;
+
+		ap->gtk_set = false;
+
+		msg = ap_build_cmd_del_key(ap);
+		if (!l_genl_family_send(nl80211, msg, ap_gtk_op_cb, NULL,
+					NULL)) {
+			l_genl_msg_unref(msg);
+			l_error("Issuing DEL_KEY failed");
+		}
+	}
+
 	ap->pending = l_dbus_message_ref(message);
 
 	return 0;
@@ -1406,11 +1639,27 @@ static struct l_dbus_message *ap_dbus_stop(struct l_dbus *dbus,
 	return NULL;
 }
 
+static bool ap_dbus_property_get_started(struct l_dbus *dbus,
+					struct l_dbus_message *message,
+					struct l_dbus_message_builder *builder,
+					void *user_data)
+{
+	struct ap_state *ap = user_data;
+	bool started = ap->started;
+
+	l_dbus_message_builder_append_basic(builder, 'b', &started);
+
+	return true;
+}
+
 static void ap_setup_interface(struct l_dbus_interface *interface)
 {
 	l_dbus_interface_method(interface, "Start", 0, ap_dbus_start, "",
 			"ss", "ssid", "wpa2_psk");
 	l_dbus_interface_method(interface, "Stop", 0, ap_dbus_stop, "", "");
+
+	l_dbus_interface_property(interface, "Started", 0, "b",
+					ap_dbus_property_get_started, NULL);
 }
 
 static void ap_destroy_interface(void *user_data)
@@ -1420,34 +1669,39 @@ static void ap_destroy_interface(void *user_data)
 	ap_free(ap);
 }
 
-static void ap_add_interface(struct device *device)
+static void ap_add_interface(struct netdev *netdev)
 {
 	struct ap_state *ap;
 
 	/* just allocate/set device, Start method will complete setup */
 	ap = l_new(struct ap_state, 1);
-	ap->device = device;
+	ap->netdev = netdev;
 
 	/* setup ap dbus interface */
 	l_dbus_object_add_interface(dbus_get_bus(),
-			device_get_path(device), IWD_AP_INTERFACE, ap);
+			netdev_get_path(netdev), IWD_AP_INTERFACE, ap);
 }
 
-static void ap_remove_interface(struct device *device)
+static void ap_remove_interface(struct netdev *netdev)
 {
 	l_dbus_object_remove_interface(dbus_get_bus(),
-			device_get_path(device), IWD_AP_INTERFACE);
+			netdev_get_path(netdev), IWD_AP_INTERFACE);
 }
 
-static void ap_device_event(struct device *device, enum device_event event,
-								void *userdata)
+static void ap_netdev_watch(struct netdev *netdev,
+				enum netdev_watch_event event, void *userdata)
 {
 	switch (event) {
-	case DEVICE_EVENT_MODE_CHANGED:
-		if (device_get_mode(device) == DEVICE_MODE_AP)
-			ap_add_interface(device);
-		else
-			ap_remove_interface(device);
+	case NETDEV_WATCH_EVENT_UP:
+	case NETDEV_WATCH_EVENT_NEW:
+		if (netdev_get_iftype(netdev) == NETDEV_IFTYPE_AP &&
+				netdev_get_is_up(netdev))
+			ap_add_interface(netdev);
+		break;
+	case NETDEV_WATCH_EVENT_DOWN:
+	case NETDEV_WATCH_EVENT_DEL:
+		ap_remove_interface(netdev);
+		break;
 	default:
 		break;
 	}
@@ -1455,10 +1709,7 @@ static void ap_device_event(struct device *device, enum device_event event,
 
 bool ap_init(struct l_genl_family *in)
 {
-	device_watch = device_watch_add(ap_device_event, NULL, NULL);
-	if (!device_watch)
-		return false;
-
+	netdev_watch = netdev_watch_add(ap_netdev_watch, NULL, NULL);
 	nl80211 = in;
 
 	return l_dbus_register_interface(dbus_get_bus(), IWD_AP_INTERFACE,
@@ -1471,7 +1722,6 @@ bool ap_init(struct l_genl_family *in)
 
 void ap_exit(void)
 {
-	device_watch_remove(device_watch);
-
+	netdev_watch_remove(netdev_watch);
 	l_dbus_unregister_interface(dbus_get_bus(), IWD_AP_INTERFACE);
 }
